import java.util.*;
import java.awt.Font;
import java.net.URL;
import java.io.File;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Point;
import java.awt.event.*;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.Highlighter;

import net.sourceforge.nite.search.*;
import net.sourceforge.nite.meta.*;
import net.sourceforge.nite.meta.impl.*;
import net.sourceforge.nite.nom.nomwrite.*;
import net.sourceforge.nite.nom.nomwrite.impl.*;
import net.sourceforge.nite.util.SearchResultTimeComparator;
import net.sourceforge.nite.gui.util.*;
import net.sourceforge.nite.query.*;
import net.sourceforge.nite.gui.textviewer.*;
import net.sourceforge.nite.gui.mediaviewer.*;
import net.sourceforge.nite.nxt.*;
import net.sourceforge.nite.nstyle.handler.*;
import net.sourceforge.nite.time.*;

import net.sourceforge.nite.nom.NOMException;
import javax.swing.plaf.metal.*;

/**
 * Switchboard interface for animacy coding
 * 
 * This interface displays the Switchboard transcription, one
 * parse per line, with an initial field on the line showing the
 * speaker of the containing turn and the id of the parse. 
 * Markables are denoted by parentheses.  The colour of the
 * bracketing is determined by the nesting level, to make it
 * easy to match up brackets visually.  There are buttons for
 * coding the animacy, animconf (confidence about the animacy 
 * coding), and anthro attribute.
 * 
 * At the command line, the call must pass the metadata file.
 * If no observation name is passed in, then the interface
 * puts up a dialogue box that allows the user to choose from
 * the full list of observations.  The call can also pass in the
 * initial font size to use.  The last command line option, -prefix,
 * controls whether or not the codes are shown in line with the
 * transcript as a prefix for each markable, with the main code
 * shown in full and suffixed by /v if "vanim" (not confident)
 * has been chosen, and further by /a if "anthro" (used anthropomorphically)
 * has been chosen.  Whether or not -prefix is chosen, if the markable
 * has not been coded for animacy, a funny dot character follows
 * the open parenthesis.  Also, when a markable is selected, the
 * current codes on the various menus will be highlighted with a
 * red ball.
 * 
 * The user is expected to animacy code the entire dialogue, selectively
 * adding vanim and anthro codes when they please.  Later we can
 * default null choice to not-vanim and not-anthro.  The uncoded 
 * buttons are so that the user can restore a markable to its 
 * original state if they decide they don't want to deal with the
 * markable just then.  Marking animacy as uncoded also restores the
 * other two attributes to uncoded.  Forward and back buttons take
 * the user systematically through all of the markables.  The interface
 * does not move forward automatically after animacy has been coded
 * as it did in previous versions because there are now further choices
 * about vanim and anthro for the user to make.
 * 
 * @author Jonathan Kilgour, Jean Carletta April 2003

 * REVISIONS:
 * JUNE 2007, revising so that it works with the expected Switchboard
 * public release that has timing data and has changed the representation
 * for this data somewhat.
 * 
 * Aug 09, a data representation change meant that the transcription no
 * longer interleaved words from different speakers, because the tool
 * was rendering one line per parse, and these are untimed.  I've changed
 * this to rendering one line per turn. (JC)
 *
 **/

public class SwitchboardAnimacy
    implements ActionListener, MouseListener, WindowListener{

    /* SOME CONSTANTS that make it easier to adapt this code */
    //the exact name of the element being marked up
    private final String MARKABLE = "markable";

    //the exact names of the attributes being marked up
    // (not all of these get used in this program; some are
    // for the Edinburgh coder)
    private final String ANIMACY = "animacy";
    private final String ANTHRO = "anthro";
    private final String ANIMCONF = "animconf";
    private final String STATUS = "status";
    private final String STATUSTYPE = "statustype";
    // what to put on the button for adding a note
    private final String NOTEBUTTON = "Add note";
    // the name of the attribute for storing notes
    private final String NOTE = "stan-note";

    // the attribute values for these attributes we wish to
    // treat as privileged, meaning "uncoded".  Uncoded means
    // that the coder hasn't considered the value yet - the attribute
    // value is actually missing if they haven't considered it, because
    // that's the way the initial data was imported, but this explicit
    // value allows them to say "pretend I haven't done anything with it".
    private final String ANIMUNCODED = "anim-uncoded";
    private final String ANTHROUNCODED = "anthro-uncoded";
    private final String ANIMCONFUNCODED = "animconf-uncoded";
    private final String STATUSUNCODED = "status-uncoded";
    private final String STATUSTYPEUNCODED = "statustype-uncoded";

    // a funny symbol to show on screen meaning "uncoded", used
    // animacy in the Stanford coding and status in the Edinburgh coding
    // (since these are the main attributes - without them, the user
    // isn't meant to make choices for the other attributes).
    private final String UNCODED_STRING = "\u00B7";

    // names of buttons for moving forward to the next markable and 
    // back to the previous one
    private final String FORWARD = "forward";
    private final String BACKWARD = "back";

    // set redisplay=false if you want to speed up the interface
    // by coding blind (no longer necessary and may no longer work)
    private final boolean redisplay = true;

    /* STORAGE FOR THE DATA */
    private NOMWriteCorpus nom;
    private NiteMetaData controlData;

    /* STORAGE FOR COMMAND LINE ARGUMENTS */
    private String corpusname;
    private String observationname;
    private String fontsize; //contains a string rep. of a number, like "12" 
    private boolean check;
    // if check is true, show codes as a prefix for each markable and forget about
    //the red ball highlighting the chosen code button, and forgetting about jumping
    // to the next markable (unless the skip button is chosen). 

    /* COMPONENTS */
    // the frame outside everything
    private JFrame frame;
    // a desktop within that frame, containing the transcription and coding windows
    private JDesktopPane desktop;
    // the component that shows the transcription.
    private JInternalFrame textIFrame;
    // the text area in that frame
    private NTextArea nta;
    // components for coding buttons.
    private JInternalFrame animIFrame; //animacy codes
    private JInternalFrame anthroIFrame; //anthropomorphism codes
    private JInternalFrame animconfIFrame; //certainty about animacy codes
    private JInternalFrame skipIFrame; // skip and back buttons
    private JInternalFrame noteIFrame; //type-in window for notes.
    private JTextField noteTextField;
    private JInternalFrame traceIFrame;

	
    /* INTERFACE MEMORY */
    // the markable being changed
    private NOMElement changed = null;
    // haven't a clue
    private NTextElement high = null;

    /* CONTROLLING HOW THE THING LOOKS */
    // fontsize command line argument will determine textsize (the fontsize
    // to use for basic text and for button labels) and bracketsize (the fontsize
    // to use for brackets, slightly larger to make them easier to click on)
    private int textsize;
    private int bracketsize;

    // global counter of markable nesting, so that we can colour them differently
    private int marknest = 0;

    // NITE text areas require one to construct styles in advance to use them.
    // We need two types for this interface:  ones for normal text, and ones
    // for brackets in a slightly larger size.  We'll keep these in two separate
    // lists.
    // The stylenamelists have two purposes:  telling the colour name to use
    // and giving a name that the text area can use to index the corresponding style.
    // Since the names of the styles need to be unique, we'll be appending "text"
    // to the styles for the normal text, and leaving stylenames from the bracket
    // list as they are.  
    private String[] normalstylenamelist = { "black" };
    private Style[] normalstylelist = new Style[normalstylenamelist.length];
    private String[] bracketstylenamelist =
    { "black", "blue", "red", "green", "yellow" };
    private Style[] bracketstylelist = new Style[bracketstylenamelist.length];

    private Stack markable_stack = new Stack();

    //a list of the markables in visitation order for the coder to use
    //to control the forward and back buttons.
    private LinkedList markable_list = new LinkedList();
    private String redball_filename = "/net/sourceforge/nite/icons/misc/ballred.gif";
    private ImageIcon highlight_image;

    /* MAPS used to control the ball indicator */
    private HashMap animbuttonmap = new HashMap();
    private HashMap animconfbuttonmap = new HashMap();
    private HashMap anthrobuttonmap = new HashMap();

    /* SEARCH engine and window for the search menu */
    private Engine searchEngine = new Engine();
    //OLD VERSION private QueryWindow search = null;
    private net.sourceforge.nite.search.GUI search = null;
    private Comparator mycomp = new SearchResultTimeComparator();

    ///////////////////////////////////////////////////
    ////////////////////////////////////////////////////
    public SwitchboardAnimacy() {
    }

    public SwitchboardAnimacy(String c, String o, String s, boolean p) {
	corpusname = c;
	observationname = o;
	fontsize = s;
	check = p;
	try {
	    controlData = new NiteMetaData(corpusname);
	    nom = new NOMWriteCorpus(controlData);
	    if (observationname == null) {
		ChooseObservation co = new ChooseObservation(controlData);
		observationname = co.popupDialog();
	    }
	    if (observationname != null) {
		NObservation obs =
		    controlData.getObservationWithName(observationname);
		nom.loadData(obs);
		System.out.println("Finished loading.");
	    }
	    //OLD VERSION
	    //search = new QueryWindow(nom);
            search = new net.sourceforge.nite.search.GUI(nom);
	    setupInterface(nom, fontsize);
	} catch (NiteMetaException nme) {
	    nme.printStackTrace();
	} catch (NOMException nex) {
	    nex.printStackTrace();
	}
    }

    private void setupInterface(NOMWriteCorpus nom, String fontsize) {
	frame = new JFrame();
	frame.addWindowListener(this);
	highlight_image = loadIcon(redball_filename);
	frame.setTitle("Switchboard Corpus Animacy Coder -- " + observationname);
	makeFileMenu(frame);
	desktop = new JDesktopPane();
	figureFontSize(fontsize);
	makeStyles(textsize, bracketsize);
	skipIFrame =
	    new JInternalFrame(
			       "Skip and back buttons",
			       true,
			       false,
			       true,
			       true);
	skipIFrame.setSize(new Dimension(200, 95));
	skipIFrame.setLocation(new Point(500, 0));
	makeSkipAndBackButtons(skipIFrame);
	desktop.add(skipIFrame);
	// first set of buttons, for animacy codes.  This is the main set
	// so it's the only one to have skip/back buttons that move the
	// interface on.
	animIFrame =
	    new JInternalFrame("Markable animacy", true, false, true, true);
	animIFrame.setSize(new Dimension(200, 295));
	animIFrame.setLocation(new Point(500, 100));
	animbuttonmap = makeCodingButtons(animIFrame, MARKABLE, ANIMACY);
	desktop.add(animIFrame);
	// second set of buttons, for certaincy about animacy codes
	animconfIFrame =
	    new JInternalFrame(
			       "Confidence about animacy",
			       true,
			       false,
			       true,
			       true);
	animconfIFrame.setSize(new Dimension(200, 95));
	animconfIFrame.setLocation(new Point(500, 400));
	animconfbuttonmap =
	    makeCodingButtons(animconfIFrame, MARKABLE, ANIMCONF);
	desktop.add(animconfIFrame);
	// third set of buttons, for whether animacy use is anthropomorphic
	anthroIFrame =
	    new JInternalFrame(
			       "Markable anthropomorphism",
			       true,
			       false,
			       true,
			       true);
	anthroIFrame.setSize(new Dimension(200, 95));
	anthroIFrame.setLocation(new Point(500, 500));
	anthrobuttonmap = makeCodingButtons(anthroIFrame, MARKABLE, ANTHRO);
	desktop.add(anthroIFrame);
		
	noteIFrame =
	    new JInternalFrame(
			       "Note",
			       true,
			       false,
			       true,
			       true);
	noteIFrame.setSize(new Dimension(500, 100));
	noteIFrame.setLocation(new Point(0, 620));
	makeNoteTypeIn(noteIFrame);
	desktop.add(noteIFrame);
		

	/* MARKABLE DISPLAY */
	textIFrame =
	    new JInternalFrame(
			       "Text Display with Markables",
			       true,
			       false,
			       true,
			       true);
	textIFrame.setSize(new Dimension(500, 600));
	addTextAreaContent(textIFrame);
	/* Find the element for initial highlight */
	if (markable_list.size() != 0) {
	    changed = (NOMElement) markable_list.getFirst();
	}
	desktop.add(textIFrame);

	desktop.setSize(new Dimension(750, 640));
	frame.getContentPane().add(desktop);
	frame.setSize(new Dimension(800, 650));
	frame.setVisible(true);

	if (changed != null) {
	    nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(changed));
	    showCodeHighlights(changed);
	    showNote(changed);
	    //showStatus(changed);
	}
    }
	

    /** Turn off the little icons next to coding buttons for
     * all the sets of buttons that use them on this interface.
     */
    private void clearCodeHighlights() {
	clearIcons(animbuttonmap);
	clearIcons(animconfbuttonmap);
	clearIcons(anthrobuttonmap);
    }

    /** Update the little icons next to coding buttons for
     * all the sets of buttons that use them on this interface.
     */
    private void showCodeHighlights(NOMElement nomel) {
	showCode(nomel, ANIMACY, ANIMUNCODED, animbuttonmap, highlight_image);
	showCode(
		 nomel,
		 ANIMCONF,
		 ANIMCONFUNCODED,
		 animconfbuttonmap,
		 highlight_image);
	showCode(
		 nomel,
		 ANTHRO,
		 ANTHROUNCODED,
		 anthrobuttonmap,
		 highlight_image);
    }

    private void showNote(NOMElement nomel) {
	String text = (String)nomel.getAttributeComparableValue(NOTE);
	noteTextField.setText(text);
    }
	
    private void clearNote() {
	noteTextField.setText("");
    }
	
    /** add a file menu.  Most of these facilities draw on things from the
     * gui.util package.
     */
    private void makeFileMenu(JFrame frame) {
	JMenuBar menubar = new JMenuBar();
	JMenu file = new JMenu("File");
	file.add(new OpenObservationAction());
	file.add(new SaveAction());
	file.add(new SaveAsAction());
	file.add(new RedisplayAction());
	file.add(new FontSizeAction());
	file.add(new PrintAction());
	file.add(new ExitAction());
	menubar.add(file);
	/* Make the search menu too! */
	JMenu search = new JMenu("Search");
	search.add(new SearchAction());
	menubar.add(search);

	frame.setJMenuBar(menubar);
    }

    /** put buttons corresponding to skip forwards and skip backwards
     *  that don't change the coding of the current markable, put
     * them on a panel on the internal frame.
     */
    private void makeSkipAndBackButtons(JInternalFrame f) {
	//JSplitPane jsp = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
	//JPanel buttonspanel = new JPanel();
	//addButtons(buttonspanel, elname, attname);
	JPanel skipbuttonpanel = new JPanel();
	addSkipButton(skipbuttonpanel);
	addBackButton(skipbuttonpanel);
	//jsp.setTopComponent(skipbuttonpanel);
	//jsp.setBottomComponent(buttonspanel);
	//	f.getContentPane().add(panel);
	//f.getContentPane().add(jsp);
	f.getContentPane().add(skipbuttonpanel);
	f.setVisible(true);

    }
	
    /** put up a notes section; when the user hits return, it adds
     * the note for the currently selected markable.
     */
    private void makeNoteTypeIn(JInternalFrame f) {
	JPanel panel = new JPanel();
	JButton nbut = new JButton(NOTEBUTTON);
	nbut.setFont(new Font("Serif", Font.PLAIN, textsize));
	nbut.addActionListener(this);
	panel.add(nbut);
	f.getContentPane().add(panel);
	noteTextField = new JTextField(20);
        noteTextField.addActionListener(this);
        panel.add(noteTextField);
	f.setVisible(true);
		
    }
	

    /** put buttons corresponding to the enumerated attribute values of the
     * named attribute and element on the internal frame, and add the frame
     * to the desktop
     */
    private HashMap makeCodingButtons(
				      JInternalFrame f,
				      String elname,
				      String attname) {
	JPanel panel = new JPanel();
	HashMap bmap = addButtons(panel, elname, attname);
	f.getContentPane().add(panel);
	f.setVisible(true);
	return bmap;
    }

    /** Add the buttons to set the chosen attribute for the chosen element,
     * drawing the values from the metadata description
     */
    private HashMap addButtons(JPanel panel, String elname, String attname) {
	NElement nell = controlData.getElementByName(elname);
	NAttribute nat = nell.getAttributeByName(attname);
	HashMap bmap = new HashMap();

	List vals = nat.getEnumeratedValues();
	Iterator nit = vals.iterator();
	while (nit.hasNext()) {
	    String val = (String) nit.next();
	    JButton nbut = new JButton(val);
	    nbut.setFont(new Font("Serif", Font.PLAIN, textsize));
	    nbut.addActionListener(this);
	    bmap.put(val, nbut);
	    panel.add(nbut);
	}
	return bmap;
    }

    /** Add an extra button on another panel in the same frame
     * as the markable coding buttons that skips forward, leaving
     * the coding exactly as it was.  (The coding interface moves on
     * to the next markable when one is coded).
     */
    private void addSkipButton(JPanel panel) {
	JButton nbut = new JButton(FORWARD);
	nbut.setFont(new Font("Serif", Font.PLAIN, textsize));
	nbut.addActionListener(this);
	panel.add(nbut);
    }

    /** Add an extra button on another panel in the same frame
     * as the markable coding buttons that skips backward, leaving
     * the coding exactly as it was.  
     */
    private void addBackButton(JPanel panel) {
	JButton nbut = new JButton(BACKWARD);
	nbut.setFont(new Font("Serif", Font.PLAIN, textsize));
	nbut.addActionListener(this);
	panel.add(nbut);
    }

    /** Clear a display element frame ready for redisplay 
     */
    private void clearFrame(JInternalFrame frame) {
	frame.setVisible(false);
	frame.getContentPane().removeAll();
    }

    /** translate the fontsize command line argument into integers for the
     * font size of the text and buttons, and another, a little larger, for the
     * brackets, storing in the globals textsize and bracketsize.
     */
    private void figureFontSize(String fontsize) {
	// before making styles, figure out what size of normal text to use, with 12 as the default.
	textsize = 12;
	if (fontsize != null) {
	    try {
		Integer fs = Integer.valueOf(fontsize);
		textsize = fs.intValue();
	    } catch (NumberFormatException e) {
		System.out.println(
				   "Can't interpret "
				   + fontsize
				   + " as an integer to use as font size!");
		e.printStackTrace();
		System.exit(0);
	    }
	}
	// make the brackets a big bigger so they're easier to click on.
	bracketsize = textsize + 4;
    }

    /** NITE expects styles for use in a text area to be made ahead.  Iterating down the colour names given in bracketstylenamelist
     * and textstylenamelist, make the styles and store them in bracketstylelist and textstylelist.  
     * Styles are made using a TextStyleHandler.  
     */
    private void makeStyles(int textsize, int bracketsize) {

	for (int i = 0; i < bracketstylenamelist.length; i++) {
	    TextStyleHandler stylehandler = new TextStyleHandler();
	    stylehandler.init("", null);
	    stylehandler.setName(bracketstylenamelist[i]);
	    stylehandler.makeNewStyle();
	    // args are style, foreground colour, background colour
	    stylehandler.setStyle(
				  stylehandler.getColours(
							  stylehandler.getStyle(),
							  bracketstylenamelist[i],
							  bracketstylenamelist[i]));

	    StyleConstants.setFontSize(stylehandler.getStyle(), bracketsize);
	    StyleConstants.setBold(stylehandler.getStyle(), true);
	    bracketstylelist[i] = stylehandler.getStyle();
	}
	for (int i = 0; i < normalstylenamelist.length; i++) {
	    TextStyleHandler stylehandler = new TextStyleHandler();
	    stylehandler.init("", null);
	    stylehandler.setName(normalstylenamelist[i] + "text");
	    stylehandler.makeNewStyle();
	    stylehandler.setStyle(
				  stylehandler.getColours(
							  stylehandler.getStyle(),
							  normalstylenamelist[i],
							  normalstylenamelist[i]));
	    StyleConstants.setFontSize(stylehandler.getStyle(), textsize);
	    StyleConstants.setBold(stylehandler.getStyle(), false);
	    normalstylelist[i] = stylehandler.getStyle();
	}
    }

    /* display or redisplay the main text area */
    private void addTextAreaContent(JInternalFrame frame) {
	if (nta != null && search != null) {
	    search.deregisterResultHandler(nta);
	}
	nta = new NTextArea();
	if (search != null) {
	    search.registerResultHandler(nta);
	}
	nta.addMouseListener(this);
	// add the styles to the text area
	for (int i = 0; i < bracketstylenamelist.length; i++) {
	    nta.addStyle(bracketstylenamelist[i], bracketstylelist[i]);
	}
	for (int i = 0; i < normalstylenamelist.length; i++) {
	    nta.addStyle(normalstylenamelist[i] + "text", normalstylelist[i]);
	}

	// create a list of all the parse elements in the loaded data.
	List elist = null;
	try {
	    //elist = searchEngine.search(nom, "($a parse)");
	    elist = searchEngine.search(nom, "($a turn)");
	} catch (Throwable e) {
	    e.printStackTrace();
	    System.exit(0);
	}
	if (elist == null || elist.size() <= 1) {
	    //System.err.println("NO parse ELEMENTS IN CORPUS!");
	    System.err.println("NO turn ELEMENTS IN CORPUS!");
	    System.exit(0);
	}

        /* remove first (dud) result */
	elist.remove(0);
        /* sort in time order */
	Collections.sort(elist, mycomp);
	for (int i = 0; i < elist.size(); i++) {
	    List l = (List) elist.get(i);
	    NOMElement nwe = (NOMElement) l.get(0);
	    // show something that uniquely identifies the parse (like the
	    // id), plus the speaker.
	    showParseField(nwe);
	    //render the words in the parse
	    showWords(nwe);
	    // render two newlines
	    NTextElement nte = new NTextElement("\n\n", "blacktext");
	    NOMObjectModelElement nome = new NOMObjectModelElement(nwe);
	    nte.setDataElement((ObjectModelElement) nome);
	    nta.addElement(nte);
	}

	JScrollPane pane = new JScrollPane(nta);
	frame.getContentPane().add(pane);
	frame.setVisible(true);
    }

    /** given a parse element, pack a text element in nta
     * that shows the speaker of the turn parent, hyphen,
     * parse id.
     */
    private void showParseField(NOMElement nwe) {
	if (nwe == null) {
	    return;
	}
	//figure out what the speaker is
	String speaker = null;
	speaker=nwe.getAgentName();
	/*
	List parentlist = nwe.getParents();
	for (int i = 0; i < parentlist.size(); i++) {
	    NOMElement parent = (NOMElement) parentlist.get(i);
	    if (parent.getName().equals("turn")) {
		speaker =
		    (String) parent.getAttributeComparableValue("speaker");
		// should terminate loop early, but don't know enough java
	    }
	};
	*/
	//get the id
	String id = nwe.getID();
	if ((speaker == null) || (id == null)) {
	    System.out.println(
			       "Found a parse missing either a speaker or an id!");
	} else {
	    // canonical way of adding a text element to an existing 
	    // text area
	    //create the text element with the desired string and style
	    NTextElement nte =
		new NTextElement(speaker + "-" + id + ": ", "red");
	    //create a NOMObjectModelElement to relate data to display
	    NOMObjectModelElement nome = new NOMObjectModelElement(nwe);
	    //do the relating
	    nte.setDataElement((ObjectModelElement) nome);
	    //add the text element to the text area
	    nta.addElement(nte);
	}
    }

    /** show the words in the corpus, recursively descending through
	the syntax layers and checking for markables at each stage. */
    private void showWords(NOMElement nwe) {
	if (nwe == null || nwe.getChildren() == null) {
	    return;
	}
	showMarkableParentsStart(nwe);
	for (Iterator kit = nwe.getChildren().iterator(); kit.hasNext();) {
	    NOMElement kid = (NOMElement) kit.next();
	    if (kid.getName().equalsIgnoreCase("word")) {
		showMarkableParentsStart(kid);
		String text = null;
		try {
		    text = (String) kid.getAttributeComparableValue("orth");
		} catch (Exception e) {
		};
		NOMObjectModelElement nome = new NOMObjectModelElement(kid);
		if (!markable_stack.empty()) {
		    NOMObjectModelElement parnome =
			(NOMObjectModelElement) markable_stack.peek();
		    nta.addElement(
				   new NTextElement(
						    text + " ",
						    "blacktext",
						    nome,
						    parnome));
		} else {
		    nta.addElement(
				   new NTextElement(
						    text + " ",
						    "blacktext",
						    nome));
		}
		showMarkableParentsEnd(kid);
	    } else if (kid.getName().equalsIgnoreCase("trace")) {
		showMarkableParentsStart(kid);
		NOMObjectModelElement nome = new NOMObjectModelElement(kid);
		//instead of just saying <trace>, put angle brackets around the label attribute
		//of the movement that has this trace filling its target role
		String tracetype = null;
		List pointerlist = kid.getPointersTo();
		if (pointerlist == null) {
		    //System.out.println(
		    //"Found a trace with nothing that points to it!");
		} else {
		    for (Iterator pit = kid.getPointersTo().iterator();
			 pit.hasNext();
			 ) {
			NOMPointer ptr = (NOMPointer) pit.next();
			if (ptr.getRole().equalsIgnoreCase("target")) {
			    tracetype =
				(String) ptr
				.getFromElement()
				.getAttributeComparableValue(
							     "label");
			}
		    };
		};
		if (tracetype == null) {
		    //System.out.println(
		    //	"Found a trace missing either a target role or the label of the pointing element!");
		    tracetype = "trace";
		};
		if (!markable_stack.empty()) {

		    NOMObjectModelElement parnome =
			(NOMObjectModelElement) markable_stack.peek();
		    nta.addElement(
				   new NTextElement(
						    "<" + tracetype + "> ",
						    "blacktext",
						    nome,
						    parnome));
		} else {
		    nta.addElement(
				   new NTextElement(
						    "<" + tracetype + "> ",
						    "blacktext",
						    nome));
		}
		showMarkableParentsEnd(kid);
	    } else {
		showWords(kid);
	    }
	}
	showMarkableParentsEnd(nwe);
    }

    /** display any open-markable elements in the appropriate colour */
    private void showMarkableParentsStart(NOMElement nwe) {
	if (nwe.getPointersTo() == null)
	    return;
	for (Iterator kit = nwe.getPointersTo().iterator(); kit.hasNext();) {
	    NOMPointer point = (NOMPointer) kit.next();
	    if (point == null)
		continue;
	    NOMElement parent = point.getFromElement();
	    if (parent.getName().equalsIgnoreCase(MARKABLE)) {
		marknest++;
		int index = marknest;
		if (marknest >= bracketstylenamelist.length) {
		    index = (marknest % bracketstylenamelist.length) + 1;
		    if (index == bracketstylenamelist.length) {
			index = 1;
		    }
		}
		String sname = bracketstylenamelist[index];
		NTextElement nte =
		    new NTextElement(statusStartString(parent), sname);
		NOMObjectModelElement nome = new NOMObjectModelElement(parent);
		nome.setDisplayedAttribute(ANIMACY);
		nte.setDataElement((ObjectModelElement) nome);
		nta.addElement(nte);
		markable_stack.push(nome);
		markable_list.add(parent);
		if (parent == changed) {
		    high = nte;
		}
	    }
	}
    }

    /* display any close-markable elements in the appropriate colour */
    private void showMarkableParentsEnd(NOMElement nwe) {
	if (nwe.getPointersTo() == null)
	    return;
	for (Iterator kit = nwe.getPointersTo().iterator(); kit.hasNext();) {
	    NOMPointer point = (NOMPointer) kit.next();
	    if (point == null)
		continue;
	    NOMElement parent = point.getFromElement();
	    if (parent.getName().equalsIgnoreCase(MARKABLE)) {
		int index = marknest;
		if (marknest >= bracketstylenamelist.length) {
		    index = (marknest % bracketstylenamelist.length) + 1;
		    if (index == bracketstylenamelist.length) {
			index = 1;
		    }
		}
		String sname = bracketstylenamelist[index];
		NTextElement nte =
		    new NTextElement(statusEndString(parent), sname);
		NOMObjectModelElement nome = new NOMObjectModelElement(parent);
		nte.setDataElement((ObjectModelElement) nome);
		nta.addElement(nte);
		nte = new NTextElement(" ", null);
		nta.addElement(nte);
		markable_stack.pop();
		marknest--;
	    }
	}
    }

    /** this is where we could select special chars to begin certain
	types of markable */
    private String statusStartString(NOMElement el) {
	String retstr = "(";
	String status = null;
	if (el.getAttribute(ANIMACY) != null) {
	    status = el.getAttribute(ANIMACY).getStringValue();
	};
	if (status == null
	    || status.equals("")
	    || status.equals(ANIMUNCODED)) {
	    retstr += UNCODED_STRING;
	} else {
	    //if in checking mode, show some codes.
	    if (check) {
		retstr += status;
		/* version for showing just a couple of characters of code */
		/* if (status!=null) {
		   if (status.length() > 1) { 
		   if (status.lastIndexOf("-")==0) { status = status.substring(0,2); } 
		   else { status = status.substring(0,1); }
		   }
		   retstr += status;
		   } */
		if (el.getAttribute(ANIMCONF) != null) {
		    status = el.getAttribute(ANIMCONF).getStringValue();
		    if (status.equals("vanim")) {
			retstr += "/v";
		    }
		}
		if (el.getAttribute(ANTHRO) != null) {
		    status = el.getAttribute(ANTHRO).getStringValue();
		    if (status.equals("anthro")) {
			retstr += "/a";
		    }
		}
	    }
	}
	return retstr + " ";
    }

    /** this is where we could select special chars to end certain
	types of markable */
    private String statusEndString(NOMElement el) {
	String retstr = ")";
	return retstr;
    }

	


    /** Start the application. */
    public static void main(String[] args) {
	String corpus = null;
	String observation = null;
	String size = null;
	boolean prefix = false;
	if (args.length < 2 || args.length > 8) {
	    usage();
	}
	for (int i = 0; i < args.length; i++) {
	    String flag = args[i];
	    if (flag.equals("-corpus") || flag.equals("-c")) {
		i++;
		if (i >= args.length) {
		    usage();
		}
		corpus = args[i];
	    } else if (flag.equals("-observation") || flag.equals("-o")) {
		i++;
		if (i >= args.length) {
		    usage();
		}
		observation = args[i];
	    } else if (flag.equals("-fontsize") || flag.equals("-f")) {
		i++;
		if (i >= args.length) {
		    usage();
		}
		size = args[i];
	    } else if (flag.equals("-prefix") || flag.equals("-p")) {
		prefix = true;
	    } else {
		usage();
	    }
	}
	if (corpus == null) {
	    usage();
	}

	SwitchboardAnimacy m =
	    new SwitchboardAnimacy(corpus, observation, size, prefix);
    }

    private static void usage() {
	System.err.println(
			   "Usage: java SwitchboardAnimacy -c metadata-filename -o observation-name -fontsize size");
	System.exit(0);
    }

    /** This handles all the button-presses etc for the interface */
    public void actionPerformed(ActionEvent ae) {
	String act = ae.getActionCommand();
	clearCodeHighlights();
	List sel = nta.getHighlightedModelElements(NTextArea.SELECTION_HIGHLIGHTS);
	if (sel == null || sel.size() == 0) {
	    System.out.println("Nothing selected in text area");
	} else {
	    Iterator sit = sel.iterator();
	    while (sit.hasNext()) {
		NOMObjectModelElement el = (NOMObjectModelElement) sit.next();
		if (el == null) {
		    continue;
		}
		if (!el.getName().equalsIgnoreCase(MARKABLE)) {
		    System.out.println("Please select a markable");
		    return;
		}
		changed = el.getElement();
		// if the press wasn't on the skip button
		if (!act.equals(FORWARD) && !act.equals(BACKWARD)  && !act.equals(NOTEBUTTON)) {
		    // Now we need to decide what attribute we're changing
		    // act, what the button returns, is a string that 
		    // has been set up to match an attribute value.  Go
		    // through the possible attribute values of the attributes
		    // being coded with this interface, finding the one it
		    // matches.
		    NElement nell = controlData.getElementByName(MARKABLE);
		    NAttribute nat = nell.getAttributeByName(ANIMCONF);
		    List vals = nat.getEnumeratedValues();
		    try {
			String oldstring = statusStartString(changed);
			if (vals.contains(act)) {
			    changed.setStringAttribute(ANIMCONF, act);
			} else {
			    nat = nell.getAttributeByName(ANTHRO);
			    vals = nat.getEnumeratedValues();
			    if (vals.contains(act)) {
				changed.setStringAttribute(ANTHRO, act);
			    } else {
				changed.setStringAttribute(ANIMACY, act);
				// There's something special to do if we
				//change animacy to anim-uncoded, and that's
				//changing anthro and animconf to the equivalent
				// uncoded values, too.  
				if (act.equals(ANIMUNCODED)) {
				    changed.setStringAttribute(
							       ANIMCONF,
							       ANIMCONFUNCODED);
				    changed.setStringAttribute(
							       ANTHRO,
							       ANTHROUNCODED);
				}
				// jump to next markable if we aren't in check mode 
				if (check == false) {
				    int ind = markable_list.indexOf(changed);
				    if (ind!=-1 && ind!=markable_list.size()-1) {
					changed = (NOMElement)markable_list.get(ind+1);
				    }
				}
			    }
			}

			String newstring = statusStartString(changed);
			//System.out.println(
			//	"AA" + oldstring + "BB vs. CC" + newstring + "DD");
			if (!oldstring.equals(newstring)) {
			    // for following call, if it fails to show
			    // an effect, you've probably got a problem with
			    // oldstring not matching the actual on-screen
			    // representation.
			    nta.redisplayElement(el, oldstring, newstring);
			}
			nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(changed));
		    } catch (NOMException nex) {
			nex.printStackTrace();
		    }
		}
		/* jump to previous markable if we hit back */
		if (act.equals(BACKWARD)) {
		    //System.out.println("Going backwards.");
		    int ind = markable_list.indexOf(changed);
		    if (ind != -1 && ind != 0) {
			changed = (NOMElement) markable_list.get(ind - 1);
		    }
		}
		/* jump to next markable if we hit skip */
		if (act.equals(FORWARD)) {
		    int ind = markable_list.indexOf(changed);
		    if (ind != -1 && ind != markable_list.size() - 1) {
			changed = (NOMElement) markable_list.get(ind + 1);
		    }
		}
		if (act.equals(NOTEBUTTON)) {
		    try {
			String note = noteTextField.getText();
			changed.setStringAttribute(NOTE, note);
		    } catch (NOMException nex) {
			nex.printStackTrace();
		    }
		}
	    }
	}

	nta.clearHighlights(NTextArea.SELECTION_HIGHLIGHTS);
	nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(changed));
	showCodeHighlights(changed);
	showNote(changed);
    }

    /** Save the NOM using the parameters set on the interface
     */
    public class SaveAction extends AbstractAction {
	public SaveAction() {
	    super("Save observation");
	}
	public void actionPerformed(ActionEvent ev) {
	    try {
		controlData.writeMetaData(corpusname);
		nom.serializeCorpusChanged();
	    } catch (Exception e) {
		e.printStackTrace();
	    }
	}
    }

    /** Save the changes to the corpus using the dialogue provided   */
    public class SaveAsAction extends AbstractAction {
	public SaveAsAction() {
	    super("Save as...");
	}
	public void actionPerformed(ActionEvent ev) {
	    CorpusSaveOptions cso = new CorpusSaveOptions(nom);
	}
    }

    /** Redisplay (in case the screen gets messed up)  */
    public class RedisplayAction extends AbstractAction {
	public RedisplayAction() {
	    super("Redisplay...");
	}
	public void actionPerformed(ActionEvent ev) {
	    List sel = nta.getHighlightedModelElements(NTextArea.SELECTION_HIGHLIGHTS);
	    if (sel != null || sel.size()!=0) {
		NOMObjectModelElement el = (NOMObjectModelElement) sel.iterator().next();
		changed=el.getElement();
	    } 
	    clearFrame(textIFrame);
	    addTextAreaContent(textIFrame);
	    nta.clearHighlights(NTextArea.SELECTION_HIGHLIGHTS);
	    if (changed!= null) {
		nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(changed));
		//showStatus(changed);
	    }
	}
    }

    /** change the font size (which also changes the size of the buttons) */
    public class FontSizeAction extends AbstractAction {
	public FontSizeAction() {
	    super("Font Size...");
	}
	public void actionPerformed(ActionEvent ev) {
	    ChooseFontSize cfs = new ChooseFontSize(textsize);
	    String fontsize = cfs.popupDialog();

	    // I think we can just create new styles overwriting the old names,
	    // but if not, need code here that nta.removeStyle(Style) for each one.
	    figureFontSize(fontsize);
	    makeStyles(textsize, bracketsize);

	    // resize the coding buttons
	    clearFrame(skipIFrame);
	    makeSkipAndBackButtons(skipIFrame);
	    clearFrame(animIFrame);
	    animbuttonmap = makeCodingButtons(animIFrame, MARKABLE, ANIMACY);
	    clearFrame(anthroIFrame);
	    anthrobuttonmap = makeCodingButtons(anthroIFrame, MARKABLE, ANTHRO);
	    clearFrame(animconfIFrame);
	    animconfbuttonmap =
		makeCodingButtons(animconfIFrame, MARKABLE, ANIMCONF);

	    clearFrame(noteIFrame);
	    makeNoteTypeIn(noteIFrame);

	    //redisplay main text area
	    List sel = nta.getHighlightedModelElements(NTextArea.SELECTION_HIGHLIGHTS);
	    if (sel != null || sel.size()!=0) {
		NOMObjectModelElement el = (NOMObjectModelElement) sel.iterator().next();
		changed=el.getElement();
	    } 
	    clearFrame(textIFrame);
	    addTextAreaContent(textIFrame);
	    nta.clearHighlights(NTextArea.SELECTION_HIGHLIGHTS);
	    if (changed!= null) {
		nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(changed));
		//showStatus(changed);
		showCodeHighlights(changed);
		showNote(changed);
	    }
	}
    }

    /** A very simple "exit" action - just checks for changes and saves if so */
    public class ExitAction extends AbstractAction {
	public ExitAction() {
	    super("Exit");
	}
	public void actionPerformed(ActionEvent ev) {
	    exit();
	}
    }

    private void exit() {
	CheckSave co = new CheckSave(nom);
	if (search != null) {
	    search.savePreferences();
	}
	if (!(co.popupDialog() == JOptionPane.CANCEL_OPTION)) {
	    System.exit(0);
	}
    }

    /** A "search" action - pops up a search GUI */
    public class SearchAction extends AbstractAction {
	public SearchAction() {
	    super("Search...");
	}
	public void actionPerformed(ActionEvent ev) {
	    search.popupSearchWindow();
	}
    }

    /** Change observation but check for changes first */
    public class OpenObservationAction extends AbstractAction {
	public OpenObservationAction() {
	    super("Open...");
	}
	public void actionPerformed(ActionEvent ev) {
	    CheckSave co = new CheckSave(nom);
	    if (!(co.popupDialog() == JOptionPane.CANCEL_OPTION)) {
		ChooseObservation cob = new ChooseObservation(controlData);
		String ob = cob.popupDialog();
		if (ob == null || ob.length() == 0)
		    return;
		observationname = ob;
		System.out.println("Opening " + observationname);
		nom.clearData();
		markable_list = new LinkedList();
		NObservation obs =
		    controlData.getObservationWithName(observationname);
		try {
		    nom.loadData(obs);
		} catch (NOMException nex) {
		    nex.printStackTrace();
		}
		clearFrame(textIFrame);
		addTextAreaContent(textIFrame);
		nta.clearHighlights(NTextArea.SELECTION_HIGHLIGHTS);
		changed = (NOMElement)markable_list.getFirst();
		nta.setHighlighted(NTextArea.SELECTION_HIGHLIGHTS, new NOMObjectModelElement(changed));
		clearCodeHighlights();
	    }
	}
    }

    /** Print text area */
    public class PrintAction extends AbstractAction {
	public PrintAction() {
	    super("Print...");
	}
	public void actionPerformed(ActionEvent ev) {
	    nta.printTextArea();
	}
    }

    private void clearIcons(HashMap bmap) {
	for (Iterator bit = bmap.values().iterator(); bit.hasNext();) {
	    JButton but = (JButton) bit.next();
	    if (but.getIcon() != null) {
		but.setIcon(null);
		but.repaint();
	    }
	}
    }

    /** paint an image icon on the button corresponding to the attribute for the 
     * current element, unless it's the uncoded value that's been made explicit
     */
    private void showCode(
			  NOMElement nel,
			  String att,
			  String uncodedval,
			  HashMap bmap,
			  ImageIcon im) {
	//System.out.println("Showing code for " + nel.getName() + " with " + att);
	String st = (String) nel.getAttributeComparableValue(att);
	if ((st == null) || (st == uncodedval)) {
	    return;
	}
	JButton mybutton = null;
	if (bmap.get(st) != null) {
	    mybutton = (JButton) bmap.get(st);
	}
	if (mybutton != null) {
	    mybutton.setIcon(im);
	    mybutton.repaint();
	}
    }


    /*
      private void showStatus(NOMElement nel) {
      if (check == false) {
      String st = (String)nel.getAttributeComparableValue(ANIMACY);
      if (st==null) { return; }
      JButton mybutton=null;
      if (buttonmap.get(st) != null) { mybutton = (JButton)buttonmap.get(st);	}
      if (mybutton!=null) {
      mybutton.setIcon(highlight_image);
      mybutton.repaint();
      }
      }
      }
    */

    ////////////////////////////////////////////////////////
    //Mouse listener implementation 
    public void mouseClicked(MouseEvent e) {

    }
    public void mouseEntered(MouseEvent e) {

    }
    public void mouseExited(MouseEvent e) {

    }
    public void mousePressed(MouseEvent e) {

    }
    public void mouseReleased(MouseEvent e) {
	List sel = nta.getHighlightedModelElements(NTextArea.SELECTION_HIGHLIGHTS);
	if (sel == null || sel.size() == 0) {
	    System.out.println("Nothing selected in text area");
	    return;
	}
	clearCodeHighlights();
	clearNote();
	for (Iterator sit = sel.iterator(); sit.hasNext();) {
	    NOMObjectModelElement el = (NOMObjectModelElement) sit.next();
	    if (el.getName().equals(MARKABLE)) {
		//System.out.println("Selection: " + el.getElement().getID());
		NOMElement nel = el.getElement();
		showCodeHighlights(nel);
		showNote(nel);
	    }
	}
    }

    private ImageIcon loadIcon(String name) {
	try {
	    // get resource
	    URL resource = ClassLoader.getSystemClassLoader().getResource(name);
	    if (resource == null) {
		resource = getClass().getResource(name);
	    }
	    // this is the line that doesn't work when running under eclipse
	    return new ImageIcon(frame.getToolkit().getImage(resource));
	    //return new ImageIcon( frame.getToolkit().getImage(name) );
	} catch (Exception e) {
	    System.err.println("WARNING: Failed to load image " + name);
	    return null;
	}
    }

    /////////////////////////////////////////////////////////////////////////////
    // WindowListener

    /** NOP (method needed for implementing WindowListener).
     * @param event parameter isn't used */
    public void windowActivated(WindowEvent event) {
    }

    /** NOP (method needed for implementing WindowListener).
     * @param event parameter isn't used */
    public void windowClosed(WindowEvent event) {
    }

    /**
     * Processes {@linkplain #exit()}.
     * @param event parameter isn't used
     */
    public void windowClosing(WindowEvent event) {
	exit();
    }

    /** NOP (method needed for implementing WindowListener).
     * @param event parameter isn't used */
    public void windowDeactivated(WindowEvent event) {
    }

    /** NOP (method needed for implementing WindowListener).
     * @param event parameter isn't used */
    public void windowDeiconified(WindowEvent event) {
    }

    /** NOP (method needed for implementing WindowListener).
     * @param event parameter isn't used */
    public void windowIconified(WindowEvent event) {
    }

    /** NOP (method needed for implementing WindowListener).
     * @param event parameter isn't used */
    public void windowOpened(WindowEvent event) {
    }

}
